/*
 * imageCode.c version 2.1 part of the Epspd88 2.1 package
 *
 * Copyright 1996 Justin Mitchell (madmitch@discordia.org.uk) and friends as Vfloppy.
 *
 * This version delivered in 2010 by Fred Jan Kraan (fjkraan@xs4all.nl)
 *
 * Epspd88 is placed under the GNU General Public License in April 2010.
 *
 *  This file is part of Epspd88 2.1.
 *
 *  Epspd88 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Epspd88 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Epspd88; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * imageCode - image functions for epspd.c.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "epspHeaders.h"
#include "epspv3.h"
#include "fdc.h"

extern int debug;
extern struct driveInfo driveInfo;

unsigned long int getSectorLocation(int track, int record)
{
        unsigned long location = SECTORBASE +
		track * TRACK * HEADS +
		(record - 1) * (RECORD) +
		SECTORHEADER + SECTORHEADER * ((record - 1)  / (SECTOR / RECORD));
        if (debug > 2)
                printf("   getSectorLocation; Track: %02X Record: %02X => location: %06lX\n",
                                            track,    record,       location);
	return location;
}

int imageRead(int fd, char *data, unsigned long block, int len)
{
        int err;
        int ct;
        if (debug > 3) printf("\n   imageRead; fd(%d, %s) block=%06lX len=%02X:  ", fd, deviceNames[fd], block, len);
        if (debug > 3)  printf("Dir.block: %02lX\n", (((block / 0x80) - 0x11) / 16 + 1) & 0xFF);
        if(lseek(fd, block, 0)==-1)
        {
                perror("\n> > > > > > imageRead: lseek < < < < < <");
                printf("\n   imageRead; fd(%d, %s) block=%06lX len=%02X:  ", fd, deviceNames[fd], block, len);
                return -1;
        }
        err = read(fd, data, len);
        if (debug > 3) {
                for(ct=0; ct < len; ct++)
                        printf("%02X ",data[ct] & 0xFF);
        }
        if (err < 0) {
                printf("    imageRead; err=%d\n", err);
        }
        if (debug > 3) printf("\n");
        return err;
}

int imageWrite(int fd, char *data, unsigned long block, int len)
{
        int err;
        int ct;
        if (debug > 3) printf("   imageWrite; fd(%d, %s) block %06lX, len %02X:  ", fd, deviceNames[fd], block, len);
        if (debug > 3)  printf("Dir.block: %02lX\n", (((block / 0x80) - 0x11) / 16 + 1) & 0xFF);
        if(lseek(fd, block, 0) == -1)
        {
                perror("\n> > > > > > imageWrite: lseek < < < < < <");
                printf("\n   imageWrite; fd(%d, %s) block=%06lX len=%02X:  ", fd, deviceNames[fd], block, len);
                return -1;
        }
        err = write(fd, data, len);
        if (debug > 3) {
                for(ct=0; ct < len; ct++)
                        printf("%02X ",data[ct] & 0xFF);
        }
        if (err < 0) {
                printf("   imageWrite; err=%d\n", err);
                return err;
        }
        if (debug > 3) printf("\n");
        return 1;
}

